/*
 * Copyright (C) 2010 Panasonic Corporation
 * All Rights Reserved.
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <common.h>
#include <command.h>
#include <malloc.h>
#include <net.h>
#include <miiphy.h>

#include <asm/arch/ethernet.h>
#if defined(AVEV4)
#include <linux/mii.h>
#endif

static int phyid = -1;	/* Phy id*/

static int rx_pos = 0;	/* Current RX Descriptor pos. */
static int rx_siz = 0;	/* RX buf size */
static int rx_off = 0;	/* RX offset */
static int tx_off = 0;	/* TX offset */
static int tx_num = 0;	/* number of txed packets */

/* used when copy needed */
static uchar  tx_adj_PacketBuf[PKTSIZE_ALIGN + PKTALIGN];
static uchar *tx_adj_buf;

static ulong mn_avev3_readreg( struct eth_device *dev,  ulong addr)
{
	ulong val;
	val = MN_AVEV3_REG(dev->iobase, addr);
	return val;
}

static void mn_avev3_writereg( struct eth_device *dev, ulong addr,  ulong val)
{
	MN_AVEV3_REG(dev->iobase, addr) = val;
}

	/* (descriptor r/w) i:index  j:0=CMDSTS, 1=BUFPTR */
static void mn_avev3_write_rxdm( struct eth_device *dev, int i, int j, ulong val)
{
	mn_avev3_writereg(dev, MN_AVEV3_RXDM + i*8 + j*4, val);
}
static void mn_avev3_write_txdm( struct eth_device *dev, int i, int j, ulong val)
{
	mn_avev3_writereg(dev, MN_AVEV3_TXDM + i*8 + j*4, val);
}

static ulong mn_avev3_read_rxdm( struct eth_device *dev, int i, int j)
{
	return mn_avev3_readreg(dev, MN_AVEV3_RXDM + i*8 + j*4);
}
static ulong mn_avev3_read_txdm( struct eth_device *dev, int i, int j)
{
	return mn_avev3_readreg(dev, MN_AVEV3_TXDM + i*8 + j*4);
}

#if defined(AVEV4)
int mn_avev3_mdio_read(struct eth_device *dev, int phy_id, int location)
{
	unsigned long mdio_ctl;
	unsigned long reg = 0;
	long          loop_cnt = 0;

	mdio_ctl = mn_avev3_readreg(dev, MN_AVEV3_MDIOCTR);

	/* Write Address Information to Address Register */
	mn_avev3_writereg(dev, MN_AVEV3_MDIOAR, (((phy_id & 31) << 8) |
              (location & 31)));

	mn_avev3_writereg(dev, MN_AVEV3_MDIOCTR, (mdio_ctl | MN_AVEV3_MDIOCTR_RREQ));

	/* Wait until completion */
	/* perhaps about 30usec...(from measured time) */
	loop_cnt = 0;
	while (1) {
		reg = mn_avev3_readreg(dev, MN_AVEV3_MDIOSR);
		if ( !(reg & MN_AVEV3_MDIOSR_STS) ) {
			break;
		}
		udelay(10); /* overhead=0.5usec. */
		loop_cnt++;
		if( loop_cnt > 1000) {
			printf("MDIO read error (MDIOSR=0x%08lx).\n", reg);
			return 0;
		}
	}

	/* Read Data */
	reg = mn_avev3_readreg(dev, MN_AVEV3_MDIORDR);

	return (int)reg;
}

static int mn_avev3_wait_autoneg_complete(struct eth_device *dev)
{
	int i;

	for(i = 0; i < 2000; i++) {
		if((mn_avev3_mdio_read(dev, phyid, MII_BMSR) & BMSR_LSTATUS) !=0) {
			break;
		}
	}

	return (i < 2000) ? 0 : -1;
}

/*
 * autoneg result check proc
 */
void mn_avev3_autoneg_result_check(struct eth_device *dev, unsigned long *speedP, unsigned long *duplexP, unsigned long *pauseP)
{
	unsigned long speed = 1000;
	unsigned long duplex = 1;
	unsigned long pause_cap = 1;
	unsigned long gb_advertise=0, gb_partner=0;
	unsigned long advertise=0, partner=0;
	unsigned long state = 0;

	/* Read GBCR, GBSR */
	gb_advertise = mn_avev3_mdio_read(dev, phyid, MII_CTRL1000);
	gb_partner = mn_avev3_mdio_read(dev, phyid, MII_STAT1000);

	/* Read ANAR, ANLPAR */
	advertise = mn_avev3_mdio_read(dev, phyid, MII_ADVERTISE);
	partner = mn_avev3_mdio_read(dev, phyid, MII_LPA);

	/*** speed, duplex status check ***/
	if( ((gb_advertise & ADVERTISE_1000FULL) != 0)
		&& ((gb_partner & LPA_1000FULL) != 0) ){
		duplex = 1;
		speed = 1000;
	} else if( ((gb_advertise & ADVERTISE_1000HALF) != 0)
		&& ((gb_partner & LPA_1000HALF) != 0) ){
		duplex = 0;
		speed = 1000;
	} else{
		state = (advertise & partner);
		if( (state & LPA_100FULL) != 0 ){
			duplex = 1;
			speed = 100;
		} else if( (state & LPA_100HALF) != 0 ){
			duplex = 0;
			speed = 100;
		} else if( (state & LPA_10FULL) != 0 ){
			duplex = 1;
			speed = 10;
		} else{
			duplex = 0;
			speed = 10;
		}
	}

	/*** pause status check ***/
	if( ((advertise & partner) & LPA_PAUSE_CAP) == 0 ){
		pause_cap = 0;
	}

	*speedP = speed;
	*duplexP = duplex;
	*pauseP = pause_cap;

	return;
}

static int mn_avev3_set_txspd(struct eth_device *dev)
{
	int ret;
	unsigned long speed = 1000;
	unsigned long duplex = 1;
	unsigned long pause_cap = 1;
	ulong txspd;
	ulong txcr;

	/* check autoneg complete & get speed */
	ret = mn_avev3_wait_autoneg_complete(dev);
	if(ret >= 0) {
		mn_avev3_autoneg_result_check(dev, &speed, &duplex, &pause_cap);
		printf("%s:INFO: autoneg complete speed = %lu\n", dev->name, speed);
	} else {
		/* use defalt speed */
		printf("%s:INFO: autoneg NOT complete speed = %lu\n", dev->name, speed);
	}

	/* select txspd */
	switch(speed) {
		case 1000: txspd = 0x00020000; break;
		case 100: txspd = 0x00010000; break;
		default: txspd = 0x00000000; break;
	}

	/* set txspd & write TXCR */
	txcr = mn_avev3_readreg(dev, MN_AVEV3_TXCR);
	txcr &= 0xfffcffff;
	txcr |= txspd;
	mn_avev3_writereg(dev, MN_AVEV3_TXCR, txcr);

	return ret;
}
#endif /* AVEV4 */

static void mn_avev3_halt(struct eth_device *dev)
{
	ulong val;
	/*-- stop Tx Rx and GRST (w/o PHY RST) --*/
	val = mn_avev3_readreg(dev, MN_AVEV3_GRR);
	if(val == 0) {	/* when GRR all cleared */
		val = mn_avev3_readreg(dev, MN_AVEV3_RXCR);
		val &= ~MN_AVEV3_RXCR_RXEN;
		mn_avev3_writereg(dev, MN_AVEV3_RXCR, val);
		mn_avev3_writereg(dev, MN_AVEV3_DESCC, MN_AVEV3_DESCC_RXDSTP);
		do {
			val = mn_avev3_readreg(dev, MN_AVEV3_DESCC);
		} while((val & MN_AVEV3_DESCC_RXDSTPST) == 0);
		mn_avev3_writereg(dev, MN_AVEV3_GRR, MN_AVEV3_GRR_GRST);
	}
#if defined(AVEV4)
	mdelay(MN_AVEV3_GRST_DELAY_MSEC);
#else
	udelay(MN_AVEV3_GRST_DELAY_USEC);
#endif
}

static void mn_avev3_reset(struct eth_device *dev)
{
#if CONFIG_RMII
	ulong val;
#endif

	mn_avev3_halt(dev);

#if defined(AVEV4)
#if CONFIG_RMII
	/* RMII register reset */
	val = mn_avev3_readreg(dev, MN_AVEV3_RSTCTL);
	val &= (~MN_AVEV3_RSTCTRL_RMIIRST);
	mn_avev3_writereg(dev, MN_AVEV3_RSTCTL, val);
#endif
	mn_avev3_writereg(dev, MN_AVEV3_GRR, MN_AVEV3_GRR_CLR);
	mdelay(MN_AVEV3_GRST_DELAY_MSEC);
#if CONFIG_RMII
	/* RMII register reset clear */
	val = mn_avev3_readreg(dev, MN_AVEV3_RSTCTL);
	val |= MN_AVEV3_RSTCTRL_RMIIRST;
	mn_avev3_writereg(dev, MN_AVEV3_RSTCTL, val);
#endif
#else /* AVEV4 */
#if CONFIG_RMII
	val = mn_avev3_readreg(dev, MN_AVEV3_RSTCTL);
	val |= 0x00010000;
	mn_avev3_writereg(dev, MN_AVEV3_RSTCTL, val);
#endif
	mn_avev3_writereg(dev, MN_AVEV3_GRR, MN_AVEV3_GRR_CLR);
#endif /* AVEV4 */

	rx_pos = 0;
	rx_off = 0;
	tx_off = 0;
	tx_num = 0;
	tx_adj_buf = &tx_adj_PacketBuf[0] + (PKTALIGN -1);
	tx_adj_buf -= (ulong)tx_adj_buf % PKTALIGN;
}


static int mn_avev3_init(struct eth_device *dev, bd_t * bd)
{
	ulong val;
	int i;

	mn_avev3_reset(dev);

	if(phyid == -1){
		for ( i = 1; i < 0x20; ++i ) {
			val = mn_avev3_mdio_read(dev, i, MII_PHYSID1);
			if((val & 0xFFFF) != 0xFFFF){
				phyid = i;
				break;
			}
		}
	}
#if __BYTE_ORDER == __BIG_ENDIAN
	val = mn_avev3_readreg(dev, 0x4124);
	val |= 0x00010000;
	mn_avev3_writereg(dev, 0x4124, val);
	val = mn_avev3_readreg(dev, 0x4160);
	val |= 0x00010000;
	mn_avev3_writereg(dev, 0x4160, val);
#endif	/*__BYTE_ORDER*/

#if defined(AVEV4)
	rx_off = 2;
	tx_off = 0;  // TX is alignment freel
	val = 0x08000000;  // for RMII setting
	mn_avev3_writereg(dev, MN_AVEV3_EMCR, val);  // for RMII setting
#else
	/* set config */
	val = MN_AVEV3_EMCR_CONF;
	mn_avev3_writereg(dev, MN_AVEV3_EMCR, val);
	if(val & MN_AVEV3_EMCR_ROFE) {
		rx_off = 2;
	}
	if(val & MN_AVEV3_EMCR_TOFE) {
		tx_off = 2;
	}
#endif
	rx_siz = (PKTSIZE_ALIGN - rx_off);

	/* set num descriptors
	 * use       1 tx descriptor and
	 *   PKTBUFSRX rx descriptors
	 */
	/* set a tx descriptor */
	mn_avev3_writereg(dev, MN_AVEV3_TXDC, MN_AVEV3_TXDC_SIZE(1));
	mn_avev3_write_txdm(dev, 0, MN_AVEV3_TXDM_CMDSTS, 0);

	/* set rx descriptors */
	mn_avev3_writereg(dev, MN_AVEV3_RXDC, MN_AVEV3_RXDC_SIZE(PKTBUFSRX));
	for(i = 0; i < PKTBUFSRX; i++) {
		val = (ulong) NetRxPackets[i];
		//MN_AVEV3_CACHE_INV_B(val, rx_siz + rx_off);
		MN_AVEV3_CACHE_FLS(val, rx_siz + rx_off);
		val = MN_AVEV3_NCADDR(val);
		mn_avev3_write_rxdm(dev, i, MN_AVEV3_RXDM_BUFPTR, val);
		mn_avev3_write_rxdm(dev, i, MN_AVEV3_RXDM_CMDSTS, rx_siz);
	}

	/* set MAC address */
	val = ((dev->enetaddr[0] << 8)|(dev->enetaddr[1]));
	mn_avev3_writereg(dev, MN_AVEV3_RXMAC2R, val);
	val = ((dev->enetaddr[2] << 24)|(dev->enetaddr[3] << 16)
	      |(dev->enetaddr[4] <<  8)|(dev->enetaddr[5]));
	mn_avev3_writereg(dev, MN_AVEV3_RXMAC1R, val);
	/* clr interrupt, set interrupt mask */
	mn_avev3_writereg(dev, MN_AVEV3_GISR, MN_AVEV3_GISR_CLR);
	mn_avev3_writereg(dev, MN_AVEV3_GIMR, MN_AVEV3_GIMR_CLR);

	/* set RXEN, Full Duplex, MTU value */
	val = (MN_AVEV3_RXCR_RXEN | MN_AVEV3_RXCR_FDUP | MN_AVEV3_RXCR_MTU);
	mn_avev3_writereg(dev, MN_AVEV3_RXCR, val);

#if defined(AVEV4)
	mn_avev3_set_txspd(dev);
#endif

	/* enable TX/RX descriptor */
	mn_avev3_writereg(dev, MN_AVEV3_DESCC, (MN_AVEV3_DESCC_RDE|MN_AVEV3_DESCC_TDE));
	return 0;
}

static int mn_avev3_send(struct eth_device *dev, void *packet, int length)
{
	ulong algn;
	ulong val;
	volatile uchar *ptr = packet;	/* use uchar pointer */

	/* adjust alignment for tx_off */
	algn = (((ulong)ptr) & 0x3);
	if(algn != tx_off) {
		MN_AVEV3_MEMCPY(tx_adj_buf + tx_off, ptr, length);
		ptr = tx_adj_buf;
	} else {
		ptr -= tx_off;
	}
	/*-- check length --*/
	if(length < MN_AVEV3_MIN_XMITSIZE) {
		volatile uchar *pad = ptr + tx_off + length;
		while(length < MN_AVEV3_MIN_XMITSIZE) {
			*pad++ = 0;
			++length;
		}
	}
	/* check ownership and wait for previous xmit done */
	do {
		val = mn_avev3_read_txdm(dev, 0, MN_AVEV3_TXDM_CMDSTS);
	} while(val & MN_AVEV3_TXDM_OWN);

	MN_AVEV3_CACHE_FLS(ptr, length + tx_off);
	/* set addr */
	val = MN_AVEV3_NCADDR((ulong)ptr);

	mn_avev3_write_txdm(dev, 0, MN_AVEV3_TXDM_BUFPTR, val);
	/* set size and kick xmit */
#if defined(AVEV4)
	val = (MN_AVEV3_TXDM_OWN | length | MN_AVEV3_TX_CMDSTS_1ST | MN_AVEV3_TX_CMDSTS_LAST);
#else
	val = (MN_AVEV3_TXDM_OWN | length);
#endif
	mn_avev3_write_txdm(dev, 0, MN_AVEV3_TXDM_CMDSTS, val);
	++tx_num;

	/* wait for xmit */
	do {
		val = mn_avev3_read_txdm(dev, 0, MN_AVEV3_TXDM_CMDSTS);
	} while(val & MN_AVEV3_TXDM_OWN);

	if((val & MN_AVEV3_TXDM_OK) == 0) {
		printf("%s:INFO:%d bad send sts=0x%lux\n", dev->name, tx_num, val);
	}

	return 0;
}

static int mn_avev3_recv(struct eth_device *dev)
{
	uchar *ptr;
	int len;
	ulong sts;

	for(;;) {
		sts = mn_avev3_read_rxdm(dev, rx_pos, MN_AVEV3_RXDM_CMDSTS);
		if((sts & MN_AVEV3_RXDM_OWN) == 0) {
			/* hardware ownership : no rec'ved packets */
			break;
		}

		ptr = NetRxPackets[rx_pos] + rx_off;
		if(sts & MN_AVEV3_RXDM_OK) {
			len = sts & MN_AVEV3_RXDM_SIZE_MASK;
			/* invalidate After DMA done */
			MN_AVEV3_CACHE_INV_A(ptr, len);
			NetReceive(ptr, len);
		} else {
			printf("%s: bad packet [%d] sts=0x%lux ptr=%p (dropped)\n",
				dev->name, rx_pos, sts, ptr);
		}

		/* invalidate Before DMA start */
		//MN_AVEV3_CACHE_INV_B(NetRxPackets[rx_pos], rx_siz + rx_off);
		MN_AVEV3_CACHE_FLS(NetRxPackets[rx_pos], rx_siz + rx_off);
		/* set hardware ownership (kick recv) */
		mn_avev3_write_rxdm(dev, rx_pos, MN_AVEV3_RXDM_CMDSTS, rx_siz);

		/* set to next rec'v position */
		if(++rx_pos >= PKTBUFSRX) {
			rx_pos = 0;
		}
	}
	return 0;
}

extern int util_mac_read(unsigned char *mac_buf);

int mn_avev3_initialize(int dev_num, int base_addr)
{
	struct eth_device *dev;
	ulong val;
	uchar enetaddr[6];

	dev = malloc(sizeof(*dev));
	if (!dev) {
		return -1;
	}
	memset(dev, 0, sizeof(*dev));

	dev->iobase = base_addr;

	/* check ID */
	val = mn_avev3_readreg(dev, MN_AVEV3_IDR);
	if(val != MN_AVEV3_IDR_VALUE) {
		printf("%s: IDR fail(base=%#x)\n", __func__, base_addr);
		free(dev);
		return -1;
	}

	/* get MAC address */
	if(0 == util_mac_read(enetaddr)){
		//printf("MAC address readed from EEPROM = %02x:%02x:%02x:%02x:%02x:%02x\n", enetaddr[0], enetaddr[1], enetaddr[2], enetaddr[3], enetaddr[4], enetaddr[5]);
		eth_setenv_enetaddr("ethaddr", enetaddr);
	}
	else{
		printf("Can't read MAC address from EEPROM\n");
		eth_getenv_enetaddr("ethaddr", enetaddr);
	}
	memcpy(dev->enetaddr, enetaddr, 6);

	dev->init = mn_avev3_init;
	dev->halt = mn_avev3_halt;
	dev->send = mn_avev3_send;
	dev->recv = mn_avev3_recv;
	sprintf(dev->name, "%s-%hu", DRIVERNAME, dev_num);

	eth_register(dev);

	return 1;
}
